// ----------------------------------------------------------------------------
//
//  Copyright (C) 2017-2018 Fons Adriaensen <fons@linuxaudio.org>
//    
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// ----------------------------------------------------------------------------


#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "abprocess.h"
#include "audiofile.h"
#include "global.h"


ABprocess::ABprocess (void) :
    _convproc (0),
    _invmic (false),
    _shnorm (SH_SN3D)
{
    reset ();
}


ABprocess::~ABprocess (void)
{
    delete _convproc;
}


// Initialise for sample rate and process block size.
//
void ABprocess::init (int sampfreq, int fragsize)
{
    _sampfreq = sampfreq;
    _fragsize = fragsize;
}


// Reset internal state (for batch processing).
//
void ABprocess::reset (void)
{
    for (int i = 0; i < 8; i++)
    {
	_hipfilt [i].reset ();
	_lfeqfilt [i].reset ();
    }
    if (_convproc) _convproc->reset ();
}


// Load configuration. The process() function must
// not be called while this one runs.
//
int ABprocess::load_config (ABconfig *C)
{
    const char *p;

    p = C->convolfilepath (_sampfreq);
    if (p == 0)	return 1;
    if (init_convproc (p))
    {
	free ((void *) p);
	return 1;
    }
    free ((void *) p);

    _lfeqconf.setpars (C->_lfeq1_f / _sampfreq,
		       powf (10.0f, 0.05f * C->_lfeq1_g), 
                       C->_lfeq1_b); 

    return 0;
}


// Process _fragsize samples.
//
void ABprocess::process (float *inp [8], float *out [9])
{
    int    i, j, m;
    float  g, *q, *opp [8];
    
    // out [6] will be the 'R' channel,
    // added after the convprocution matrix.
    for (j = 0; j < 6; j++) opp [j] = out [j];
    opp [6] = out [7];
    opp [7] = out [8];
    
    // High pass and lfeq filters.
    _hipconf.process ();
    for (j = 0; j < 8; j++)
    {
        _hipfilt [j].process (&_hipconf, _fragsize, inp [j], opp [j]);
	_lfeqfilt [j].process (&_lfeqconf, _fragsize, opp [j], opp [j]);
    }

    // Apply convolution matrix.
    _convproc->process (opp, opp);

    // The 'R' channel is silent.
    memset (out [6], 0, _fragsize * sizeof (float));
    
    // Take care of inverted mic and output format.
    m = _invmic ? 0x96 : 0;
    for (j = 0; j < 9; j++)
    {
        switch (_shnorm)
        {
        case SH_N3D:
	    g = g_n3d [j];
            break;
        case SH_FUMA:
	    g = g_fuma [j];
            break;
        default: // SN3D
	    g = 1.0f;
	}
	if (m & 1) g = -g;
	m >>= 1;
	if (g == 1.0f) continue;
    	q = out [j];
	for (i = 0; i < _fragsize; i++) q [i] *= g;
    }
}


// Create and initialise the convolution processor.
//
int ABprocess::init_convproc (const char *convfile)
{
    int        inp, out, implen;
    Audiofile  A;
    float      *buff;

    delete _convproc;
    _convproc = 0;

    if (A.open_read (convfile)) return 1;
    if (   (A.chan () != 8)
        || (A.rate () != _sampfreq)
	|| (A.size () & 7))
    {
	A.close ();
	return 1;
    }
	
    implen = A.size () / 8;
    buff = new float [8 * implen];
    _convproc = new Convol88 (implen, _fragsize);
    
    for (inp = 0; inp < 8; inp++)
    {
        A.read (buff, implen);
        for (out = 0; out < 8; out++)
        {
	    _convproc->setimp (inp, out, 1.0f, buff + out, implen, 8);
        }
    }
	
    A.close ();
    delete[] buff;
    return 0;
}
