// ----------------------------------------------------------------------------
//
//  Copyright (C) 2017 Fons Adriaensen <fons@linuxaudio.org>
//    
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// ----------------------------------------------------------------------------


#ifndef __CONVOL88_H
#define __CONVOL88_H


#include <fftw3.h>


// ----------------------------------------------------------------------------


// Frequency domain data used by Convol88.
//
class Convfdata
{
private:

    friend class Convol88;

    Convfdata (int npart, int fsize);
    ~Convfdata (void);

    void clear (void);

    int              _npart;
    int              _fsize;
    fftwf_complex  **_freq_data;   
};


// ----------------------------------------------------------------------------


// 8 by 8 convolution matrix processor, using a single
// parition size equal to the 'period' argument of the
// constructor.
// Processing delay is zero, CPU load is proportional to
// the number of partitions which is 'maxlen' / 'period'
// rounded upwards.
//
class Convol88
{
public:

    enum
    {
        NINP = 8,
        NOUT = 8,
        MAXLEN = 16384,
        MAXPER = 4096
    };

    Convol88 (int maxlen, int period);
    ~Convol88 (void);

    // Set IR for input and output. 
    int  setimp (int inp, int out, float gain, float *data, int size, int step = 1);

    // Clear all IR data.
    void clear (void);

    // Reset internal state (for batch processing).
    void reset (void);

    // Process 'period' samples.
    void process (float *inp [NINP], float *out [NOUT]);

private:

    void init (void);
    void fini (void);

    int              _maxlen;                 // maximum impulse response size
    int              _period;                 // number of samples for each process()
    int              _npart;                  // number of partitions
    int              _ipart;                  // cyclic partition counter
    float           *_outbuff [NOUT];         // overlap buffers
    float           *_time_data;              // workspace
    fftwf_complex   *_freq_data;              // workspace
    fftwf_plan       _plan_r2c;               // FFTW plan, forward FFT
    fftwf_plan       _plan_c2r;               // FFTW plan, inverse FFT
    Convfdata       *_inp_fdata [NINP];       // frequency domain data 
    Convfdata       *_mac_fdata [NINP][NOUT]; // frequency domain data
};


// ----------------------------------------------------------------------------


#endif


